"""
Created on March 29, 2018

@author: grjohnson
"""

import logging
import os
import pyshark
import arrow
import copy
import pyparsing as pyp
import itertools
from lqmt.lqm.tool import PullTool
from idstools import rule
from lqmt.lqm.exceptions import ConfigurationError
from zipfile import ZipFile


class FromSnort(PullTool):
    def __init__(self, config):
        """
        ToToolName description
        :param config: configuration file
        """
        super().__init__(config)
        self._logger = logging.getLogger("LQMT.FromSnort.{0}".format(self.getName()))

    def initialize(self):
        super().initialize()

    def __check_existence(self):
        """
        Checks the configured alert and pcap paths as well as the rule file for existence.

        :return: Boolean True/False indicator configured directory/file existence
        """
        # checks each directory in the configured snort alert paths
        for i in self._config.alert_paths:
            if not os.path.isdir(i):
                self._logger.error("Exiting - alert directory doesn't exists.")
                return False

        # checks each directory in the configured snort pcap capture paths
        for x in self._config.packets_paths:
            if not os.path.isdir(x):
                self._logger.error("Exiting - alert directory doesn't exists.")
                return False

        # checks each rules file in the configured rules list
        for y in self._config.rule_list:
            if not os.path.isfile(y):
                self._logger.error("Exiting - rule file(s) don't exist.")
                return False

        return True

    def __calculate_age_threshold(self):
        """
        Calculates the shifted date/time based on UTC now.  To be used for comparing file modified times.

        :return shift: Arrow object that represents now - the configured max age.
        """
        now = arrow.utcnow()
        second = ['s', 'sec', 'secs', 'second', 'seconds']
        minute = ['m', 'min', 'minute', 'minutes']
        hour = ['h', 'hr', 'hrs', 'hour', 'hours']
        day = ['d', 'day', 'days']
        week = ['w', 'week', 'weeks']
        month = ['mon', 'month', 'months']
        year = ['y', 'yr', 'yrs', 'year', 'years']

        cols = self._config.data_age.split(' ')
        if cols[0].isdigit():
            num = -1 * int(cols[0])
        else:
            raise ConfigurationError(
                "Unable to parse file age \"{0}\" expecting %d %s format (e.g. 2 weeks)".format(self._config.data_age))

        if cols[1] in second:
            shift = now.shift(seconds=num)
        elif cols[1] in minute:
            shift = now.shift(minutes=num)
        elif cols[1] in hour:
            shift = now.shift(hours=num)
        elif cols[1] in day:
            shift = now.shift(days=num)
        elif cols[1] in week:
            shift = now.shift(weeks=num)
        elif cols[1] in month:
            shift = now.shift(months=num)
        elif cols[1] in year:
            shift = now.shift(years=num)
        else:
            raise ConfigurationError(
                "Unable to determine time string in \"{0}\". Refer to documentation for accepted time strings.".format(self._config.data_age))

        return shift

    def __check_new_content(self, path_list):
        """
        Iterates through list of file paths/names to determine which have been modified since the configured max age.

        :param path_list: List of file paths and file names to check.
        :return: Set - contains file path/names for files modified more recent than configured maximum age.
        """
        ret = set()

        for i in path_list:
            for root, dirs, fnames in os.walk(i):
                for fname in fnames:
                    filepath = os.path.join(root, fname)
                    file_age = arrow.get(os.path.getmtime(filepath))
                    shift = self.__calculate_age_threshold()
                    if file_age > shift:
                        ret.add(filepath)

        return ret

    def __full_log_parse(self, logfile):
        """
        Code utilizes py parser to set up regular expressions for a Snort Alert File (full logging only).
        It is important to note this may need adjustment based on logging style chosen and fields for entries.

        :param logfile: Snort alert log file - from full logging configuration.
        :return ret: Dictionary containing parsed alert log entries.
        """
        i = 0
        ret = {}

        integer = pyp.Word(pyp.nums)
        ip_addr = (pyp.Combine(integer+'.'+integer+'.'+integer+'.'+integer)
                   + pyp.Suppress(pyp.Optional(":" + integer)))

        header = (pyp.Suppress("[**] [")
                  + pyp.Combine(integer + ":" + integer + ":" + integer)
                  + pyp.Suppress(pyp.SkipTo("[**]", include=True)))
        cls = (
            pyp.Suppress(pyp.Optional(pyp.Literal("[Classification:")))
            + pyp.Regex("[^]]*") + pyp.Suppress(']'))

        pri = pyp.Suppress("[Priority:") + integer + pyp.Suppress("]")
        date = pyp.Combine(
            integer+"/"+integer+'-'+integer+':'+integer+':'+integer+'.'+integer)
        src_ip = ip_addr + pyp.Suppress("->")
        dest_ip = ip_addr

        # TODO: In example, classification doesn't exist, but if inconsistent output won't parse right
        # TODO: Suppressed port information from extraction for now (not in ICMP entries)
        # TODO: might want to cascade matching -> on header, then match on IPs, then match on ports
        # bnf = header+cls+pri+date+src_ip+dest_ip
        bnf = header + pri + date + src_ip + dest_ip

        try:
            with open(logfile) as snort_logfile:
                for has_content, grp in itertools.groupby(
                        snort_logfile, key=lambda x: bool(x.strip())):
                    if has_content:
                        tmpStr = ''.join(grp)
                        fields = bnf.searchString(tmpStr)
                        if fields:
                            ret[i] = {}
                            ret[i]['file'] = logfile
                            ret[i]['text'] = tmpStr
                            ret[i]['matches'] = fields
                            i += 1
        except Exception as e:
            self._logger.warning("Unable to read file {0}. Exceptions={1}".format(logfile, e))

        return ret

    def __load_alert_ids(self):
        """
        Parses each rule file to extract the relevant Identifiers, actions, headers and pre-build a matching string.

        :return rules: Dictionary containing parsed Snort rules from configured rules files.
        """
        rules = {}

        for i in self._config.rule_list:
            # uses IDS tools library to extract rules
            rule_file = rule.parse_file(i)
            for x in rule_file:
                if x.enabled:
                    rules[x.sid] = {}
                    rules[x.sid]['sid'] = x.sid
                    rules[x.sid]['gid'] = x.gid
                    rules[x.sid]['rev'] = x.rev
                    rules[x.sid]['logmatch'] = "{0}:{1}:{2}".format(str(x.gid), str(x.sid), str(x.rev))
                    rules[x.sid]['action'] = x.action
                    rules[x.sid]['header'] = x.header

        self._logger.info("Extracted {0} rules from the Snort configuration.".format(str(len(rules))))

        return rules

    def __check_content_matches(self, parsed_log, snort_ids):
        """
        Compares the parsed Snort alert log field to identify matching Snort IDs.
        The full text of the alert entry and the SRC/DST IP addresses are then returned.

        :param parsed_log: Dictionary with each Snort alert log entry parsed for comparison.
        :param snort_ids: Dictionary with parsed Snort rule entries.
        :return: Dictionary containing final matched alert log entries and SRC/DST IP addresses.
        """
        ret_text = []
        ips = set()
        ret = {}

        for key, value in parsed_log.items():
            for x, y in snort_ids.items():
                # checks the gid:sid:rev for each rule against the parsed alert entry [0][0] with IDs
                if y['logmatch'] in value['matches'][0][0]:
                    ret_text.append(value['text'])
                    ips.add(value['matches'][0][-1])  # DST IP is in last list position
                    ips.add(value['matches'][0][-2])  # SRC IP is in second to last list position

        # full text is needed for writing non-parsed context into temporary files
        if ret_text:
            ret['text'] = ret_text

        # currently IPs used as an OR and not dependent on SRC/DST so returned as a unique list of all
        if ips:
            ret['ips'] = ips

        self._logger.info(
            "Found {0} matching alert entries and extracted {1} IP addresses".format(str(len(ret_text)), str(len(ips))))
        return ret

    def __find_file_matches(self, path_list, snort_ids):
        """
        Parses a list of alert files to identify new alert entries that match desired snort id's.
        New files are determined by comparing modified time to configured maximum age.

        :param path_list: List of alert files to check for new matches
        :param snort_ids: Parsed Snort rules that contains ID's for alert file matching
        :return: Filepaths, IP Addresses - Sets containing matching files and extracted IP addresses
        """
        ret_files = set()
        file_set = set()
        ip_set = set()
        prsd_log = {}

        # create a list of files that have new entries
        file_set = self.__check_new_content(path_list)
        for filepath in file_set:
            # Parses a snort alert log file to extract fields
            prsd_log = self.__full_log_parse(filepath)

            # compare the parsed alert log entries to match desired Snort IDs
            ret = self.__check_content_matches(prsd_log, snort_ids)

            # this allows a temporary alert file to be written that only contains full text for Snort ID matches
            if 'text' in ret:
                fname = os.path.basename(filepath)
                with open(self._config.result_path + '/' + fname, 'w') as f:
                    for lines in ret['text']:
                        f.write(lines + '\n')
                ret_files.add(self._config.result_path + '/' + fname)

            # SRC/DST IP addresses are extracted from the Snort log for future PCAP matching
            if 'ips' in ret:
                ip_set.update(ret['ips'])

        return ret_files, ip_set

    def __find_pcap_matches(self, path_list, ip_match):
        """
        Uses PyShark to filter PCAPs for packets that match passed in IP addresses.
        Upon completion temporary files are written that contain only matching packets.

        :param path_list: List containing the pcap files selected for comparing.
        :param ip_match: Set containing IP addresses to filter in the PCAPs.
        :return: ret - Set containing filepaths of temporarily written filtered pcap files.
        """
        ret = set()

        if not ip_match:
            return ret

        # TODO: for now only matches IP addresses, results in extra packets being grabbed since no protocol matching
        # Pop the first entry since it is a set
        filter_str = 'ip.addr==' + str(ip_match.pop())

        # if there are more IPs in the set, add to filter string as an OR
        for x in ip_match:
            filter_str += ' or ip.addr==' + str(x)

        # returns a list of only files modified newer than the maximum configured age
        file_set = self.__check_new_content(path_list)
        for filepath in file_set:
            # creates the temporary filename, %f used to add MS granularity
            now = arrow.utcnow()
            fname = self._config.result_path + '/' + now.strftime('%Y%m%d-%H%M%S%f') + '_filtered.pcap'

            try:
                # utilizing display_filter ensures only matching packets, output_file automatically creates a new file
                pcap = pyshark.FileCapture(filepath,
                                           display_filter=filter_str,
                                           output_file=fname)

                # must load the packet matches to get output_file to write
                pcap.load_packets()
                if len(pcap) > 0:
                    ret.add(fname)  # if matches found, add temp filename for future writing results
                    self._logger.info("Successfully matched {0} packets from {1}".format(str(len(pcap)), filepath))
                else:  # pyshark creates file even if empty so remove
                    os.remove(fname)
                    self._logger.info("No matching packets from {0}".format(filepath))
                pcap.close()
            except Exception as e:
                self._logger.warning("Exception loading {0} as PCAP = {1}".format(filepath, e))

        return ret

    def process(self, alert=None):
        """
        Process function. Handles the processing of data for the tool.
        """
        file_set = set()
        ip_set = set()
        pcap_set = set()
        ret = {}
        file_list = []

        if not self.__check_existence():
            # failure logged in function
            return

        # TODO: only supports "full" logging mode for snort
        # TODO: fast, unified2, etc. unsupported

        if self._config.mode == 'full':
            # zip files with modifications more recent than configured

            # combines the alert and pcap paths
            file_list = copy.deepcopy(self._config.alert_paths)
            file_list.extend(self._config.packets_paths)

            # checks each directory to find the "new" files into a set
            file_set = self.__check_new_content(file_list)

            # write all selected files to a zip file, don't remove originals
            self.write_results(file_set, rm_file=False)

        elif self._config.mode == 'match':
            # match mode parses snort alert files and pcaps to match snort ids and IP address packets

            # read the updating rule file to extract rule context
            ret = self.__load_alert_ids()

            # parses through the snort alert logs to match snort id and return list of associated SRC/DST IP addresses
            # following this, snort alert matches are in new files in the configured result path
            file_set, ip_set = self.__find_file_matches(self._config.alert_paths, ret)

            # parse each pcap file to identify packets with IP address matches
            # following this, pcap packet matches are in new files in the configured result path
            pcap_set = self.__find_pcap_matches(self._config.packets_paths, ip_set)

            # zip all temporary alert and pcap match files and remove the temporary files
            file_set.update(pcap_set)
            self.write_results(file_set, rm_file=True)

        else:  # Unknown capture mode
            self._logger.error("Unknown configured mode = {}".format(self._config.mode))

    def commit(self):
        pass

    def cleanup(self):
        pass

    def write_results(self, file_set, rm_file=False):
        """
        Writes the final context matches to a zip file.  To be used to create an upload file.

        :param file_set: The final list of files that have content to be zipped for sharing.
        :param rm_file: Boolean flag to delete the original file after zipping.
        :return: None
        """
        try:
            if not file_set:
                self._logger.info("No Snort files identified")
                return

            # create file name based on date/time with %f used to add MS
            now = arrow.utcnow()
            fname = now.strftime('%Y%m%d-%H%M%S%f') + '_snort_logs.zip'
            filename = self._config.result_path + '/' + fname

            # create a zip file from the passed in list of files
            with ZipFile(filename, 'w') as zip:
                for x in file_set:
                    # TODO: currently does not maintain original sub folders, but flattens the zip contents
                    # TODO: with conflicting name, ZipFile appears to force write
                    no_path = os.path.basename(x)  # this removes the subfolders from the name
                    zip.write(x, arcname=no_path)

            self._logger.info(
                "Successfully matched {0} files. Zip output created at {1}".format(str(len(file_set)), filename))

            # allows users to write temporary files and them remove after successful zip
            if rm_file:
                for y in file_set:
                    os.remove(y)
                self._logger.info("Successfully deleted original files")

            # TODO: there is a 15 MB CFM upload limit, handle that at all?
        except Exception as e:
            self._logger.error("Unable to create zipped results. Exception={0}".format(e))



INPUT1 = """{ 
    "SendingSite": "SITE1", 
    "SentTimestamp": "1502815271", 
    "PayloadType": "Alert", 
    "PayloadFormat": "STIX", 
    "DataSensitivity": "noSensitivity", 
    "SharingRestrictions": "WHITE", 
    "ReconPolicy": "noTouch"
}"""


TESTCONFIG1 = """
[[Source.Filters]]
    site_includes = [ 'SiTE1' ]
    site_excludes = [ 'SitE2' ]
    payload_types = [ 'alert' ]
    payload_formats = [ 'STix' ]
    sensitivities = [ 'nosensITIvity' ]
    restrictions = [ 'whIte' ]
    reconnaissance = [ 'notouch' ]
    max_file_age = "2 mon"
"""


from .whitelist import Whitelist
from netaddr import IPNetwork,IPSet

class IPv4SubnetWL(Whitelist):
    """IPv4 Subnet whitelist.
    This class stores its whitelisted IPv4 subnets in the ipv4sn table.
    This stores the CIDR and the max and min ip specified by the CIDR.
    """
    @staticmethod
    def getCreateTable():
        """Create the table to hold the subnet information"""
        return "create table ipv4sn (cidr char(20), minip integer(11), maxip integer(11) )"

    @staticmethod
    def storeDB(conn,subnets):
        """Overwrite the subnet info in the table with the specified subnets (string) in the database."""
        c=conn.cursor()
        vals=[]
        for sn in subnets:
            cidr=IPNetwork(sn);
            vals.append( (str(cidr),cidr.first,cidr.last))
        c.execute("delete from ipv4sn")
        c.executemany("insert into ipv4sn(cidr,minip,maxip) values (?,?,?)",iter(vals))
        c.close()

    def isWhitelisted(self,conn,indicatorType,indicator):
        """Return whether or not the indicator of type indicatorType is whitelisted by this whitelist.
        If the indicator is a single address, it is whitelisted if it is included in any CIDR.
        If the indicator is a network spec, it is whitelisted if all of the addresses it represents are included in any CIDR
        """
        sn=IPNetwork(indicator)
        minip=sn.first
        maxip=sn.last

        c=conn.cursor()
        c.execute("select cidr,minip,maxip from ipv4sn where ? between minip and maxip or ? between minip and maxip order by minip",(minip,maxip))
        ipset=None
        rec=c.fetchone()
        # create a set of all ips contained network specs that contain the min and max ip specified by the indicator
        while (rec != None):
            if(ipset==None):
                ipset = IPSet(IPNetwork(rec[0]))
            else:
                ipset = ipset | IPSet(IPNetwork(rec[0]))
            rec=c.fetchone()
        # if the resulting set is empty, the indicator is not whitelisted
        if(ipset == None):
            return False
        # if the set of IPs represented by the indicator is a subset of the IPs set created above, then it is whitelisted
        ips=IPSet(sn)
        if(ips.issubset(ipset)):
            rv=True
        else:
            rv=False
        c.close()
        return rv



import logging
from lqmt.lqm.tool import ToolConfig


class SnortConfig(ToolConfig):
    """
    Configuration class for Snort Tool
    """

    def __init__(self, configData, snortToolInfo, unhandledCSV):
        """
        Constructor
        :param configData:
        :param snortToolInfo:
        :param unhandledCSV:
        """
        super().__init__(configData, snortToolInfo, unhandledCSV)
        self.logger = logging.getLogger("LQMT.ToSnort.{0}".format(self.getName()))

        modes = ['full', 'append']

        self.snort_cfg_filepaths = self.validation('config_paths', list, required=True, default=['/etc/nsm'])
        self.snort_cfg_filename = self.validation('config_filename', str, default='snort.conf')
        self.rule_paths = self.validation('rule_paths', list, required=True, default=['/etc/nsm/rules'])
        self.rule_filename = self.validation('rule_filename', str, default='stix.rules')
        self.max_rules = self.validation('max_rules_count', int, default=-1)
        self.mode = self.validation('mode', str, default='full')
        if self.mode not in modes:
            self.mode = 'full'
        self.compatible_types = ['StixFile', 'RuleFile']



class Source(object):
    """The superclass for all sources"""
    
    def getFilesToProcess(self):
        """Return an iterator that will return alert/metadta file tuples"""
        NotImplementedError

    def processed(self,datafile):
        """A file pair has been processed"""
        NotImplementedError
        
    def logStatistics(self, numAlerts):
        """Print any statistics to loggers.  This is called when the source has been completely traversed"""
        NotImplementedError


from lqmt.lqm.tool import Tool
import logging
from lqmt.lqm.data import AlertAction
import os


class ToCSV(Tool):
    def __init__(self, config):
        super().__init__(config, [AlertAction.get('All')])
        self._logger = logging.getLogger("LQMT.CSV.{0}".format(self.getName()))
        self._fp = None

    def initialize(self):
        super().initialize()
        self._fp = None

    def fileBegin(self):
        pass

    def fileDone(self):
        pass

    # Will only be called for alerts this tool can process
    def process(self, alert):
        if self._fp is None and not self._openFile():
            self.disable()
        if self.isEnabled():
            # write the alert to the file
            self._fp.write(",".join(alert.getFields(self._config.fields)) + "\n")
            self._fp.flush()
        return alert

    def commit(self):
        pass

    def cleanup(self):
        if self._fp is not None:
            self._fp.close()

    def _openFile(self):
        try:
            file = self._config.getFile()
            fileDirectory = os.path.dirname(file)
            if not os.path.exists(fileDirectory):
                os.makedirs(fileDirectory, 0o755, True)
            self._fp = open(file, 'a')
            header = ','.join(self._config.fields) + "\n"
            self._fp.write(header)

            return True

        except Exception as e:
            self._logger.error("Unable to open csv file: {0}".format(file))
            self._logger.error(e)
            return False



"""
Created on March 13, 2018

@author: grjohnson
"""
import logging
import os
import copy
from lqmt.lqm.data import AlertAction
from lqmt.lqm.tool import Tool


class ToSnort(Tool):
    def __init__(self, config):
        """
        ToSnort tool.  Used to insert Snort rules into a Snort configuration.
        :param config: configuration file
        """
        super().__init__(config, [AlertAction.get('OtherAction')])
        self._logger = logging.getLogger("LQMT.ToSnort.{0}".format(self.getName()))

    def initialize(self):
        super().initialize()

    def __setup_snort_configs(self):
        """
        Sets up the configuration files for the Snort installation.
        First step is to confirm that the snort configuration file exists.
        Next step is to confirm/create the configured rules files exist.
        Finally, if the rules files are not included in the snort configuration file - add it.
        :return Boolean: True is Snort configs are correct.  False if configs don't exist at configured path.
        """
        rule_files = []
        rule_needed = []

        # check for snort.conf existence
        for i in self._config.snort_cfg_filepaths:
            file_path = i + '/' + self._config.snort_cfg_filename
            if not os.path.isfile(file_path):
                self._logger.error(
                    "Exiting Snort configuration file doesn't exist {0}: No rules written".format(file_path))
                return False

        # check if rules file exists
        for i in self._config.rule_paths:
            file_path = i + '/' + self._config.rule_filename
            if not os.path.isfile(file_path):
                # if not create it
                open(file_path, 'w').close()
            rule_files.append(file_path)

        # check if rules file exists in the snort.conf
        for i in self._config.snort_cfg_filepaths:
            rule_needed = copy.deepcopy(rule_files)
            file_path = i + '/' + self._config.snort_cfg_filename
            with open(file_path, "r") as conf_file:
                for line in conf_file:
                    for x in rule_files:
                        if 'include ' + x in line:
                            rule_needed.remove(x)
                        if not rule_needed:  # speed this up by quitting once all are found
                            break

            # TODO: find section 7 for site specific rules or add to local rules only
            if rule_needed:
                with open(file_path, "a") as conf_file:
                    conf_file.write('\n\n')
                    for x in rule_needed:
                        self._logger.info("Added rule file to Snort configuration " + x)
                        conf_file.write('# Cyber Fed Model added rules file\n')
                        conf_file.write('include ' + x + '\n')

        return True

    def process(self, alerts):
        """
        Process function.  Handles the processing of data for the tool.
        Once completed, the rules file(s) contain the new alerts, have unique entries,
        and is smaller than the maximum count if configured to check.

        Important pre-conditions:
        - Currently only supports StixFile data objects from the Parser
        - The rules dictionary from the parser must have a 'snort' entry

        :param alerts:
        :return None:
        """
        # TODO: STIX Snort Specification allows -- Event Filters / Rate Filters / Event Suppression
        file_lines = []
        ap_char = []

        # compare to list of compatible objects
        if type(alerts).__name__ not in self._config.compatible_types:
            self._logger.warning("Unsupported Object Type {0}: No rules written".format(type(alerts).__name__))
            return

        # should confirm that snort is in rules dictionary
        snort_rules = alerts.getRules('snort')  # Requires the user to have configured parser for snort
        if not snort_rules:
            self._logger.warning("Snort rules do not exist in data object.")
            return

        try:
            x = self.__setup_snort_configs()
            if not x:
                # logging is performed in the function call
                return

            # Processing loop for each configured rules file
            for i in self._config.rule_paths:
                file_path = i + '/' + self._config.rule_filename

                if self._config.mode == 'full':
                    with open(file_path, 'w') as rule_file:
                        rule_file.write("".join(snort_rules))

                    # when in full overwrite mode, configured setting is ignored
                    if len(snort_rules) > self._config.max_rules:
                        self._logger.warning("Maximum rules configuration in ignored in Full mode.")

                    self._logger.info(
                        "Added {0} rule entries to {1}".format(str(len(snort_rules)), file_path))

                elif self._config.mode == 'append':
                    # READ: existing file for processing
                    with open(file_path, 'r') as rule_file:
                        file_lines = rule_file.readlines()

                    # this is only applicable if coming from STIX and no '\n'
                    # add \n to each new rule to compare to existing rules
                    # ap_char = list(x + '\n' for x in snort_rules)
                    for x in snort_rules:
                        if x[-1] != '\n':
                            ap_char.append(x+'\n')
                        else:
                            ap_char.append(x)

                    # utilize set to create unique list of new rules and existing rules
                    new_rule_list = set(ap_char)
                    curr_file_list = set(file_lines)

                    # MAXIMUM: check total number of new rules against the maximum allowed
                    if self._config.max_rules != -1:
                        # calculate if a the new combined set of unique entries is > maximum
                        new_file_list = copy.deepcopy(curr_file_list)
                        new_file_list.update(new_rule_list)
                        new_rules = len(new_file_list)
                        if new_rules > self._config.max_rules:
                            # if combination is too large, remove first entries to create new list
                            new_start = new_rules - self._config.max_rules
                            self._logger.info(
                                "Maximum Rules in the file, removing {0} entries".format(str(new_start)))
                            temp = list(curr_file_list)[new_start:]
                            temp2 = set(temp)
                            temp2.update(new_rule_list)
                            new_file = list(temp2)
                        else:
                            # important - if under threshold use the new list to write to file
                            new_file = list(new_file_list)
                    else:
                        # if not checking # of rules, create new list and set to write to file
                        curr_file_list.update(new_rule_list)
                        new_file = list(curr_file_list)

                    # TODO: Need to test and confirm no duplicate entries, max rules, etc.
                    # WRITE: update the rules file with remaining new rules
                    with open(file_path, 'w') as rule_file:
                        rule_file.write("".join(new_file))

                    self._logger.info(
                        "Added {0} rule entries to {1}".format(str(len(new_rule_list)), file_path))
                else:
                    self._logger.warning(
                        "No rules added to {0}. Unknown mode={1}".format(file_path, self._config.mode))

            # TODO: handle the auto restart snort command?
        except Exception as e:
            self._logger.error("Error while updating Snort Rules. Exception={0}".format(e))

    def commit(self):
        pass

    def cleanup(self):
        pass



USERCONFIG = """
[[Source.Directory]]
    dirs = [ "{0}" ]
    post_process = "nothing"

[[Source.Filters]]
    site_includes = [ ]
    site_excludes = [ ]
    payload_types = [ ]
    payload_formats = [ ]
    sensitivities = [ ]
    restrictions = [ ]
    reconnaissance = [ ]
    max_file_age = "2 mon"

[Logging]
  logfilebase = "{1}"
  debug = true

[Whitelist]
    whitelist = "{2}"
    dbfile = "{3}"

#Flextext configuration
[[Tools.FlexText]]
    name                = "flextext-tool"
    fileParser          = "CSV"
    fields              = ['indicator', 'reportedTime', 'detectedTime', 'duration1', 'priors', 'directSource', 'reason1', 'majorTags', 'sensitivity', 'reconAllowed', 'restriction']
    delimiter           = " "
    quoteChar           = "'"
    escapeChar          = '\\'
    headerLine          = true
    doubleQuote         = false
    quoteStyle          = "none"
    primarySchemaConfig = "resources/schemaDefinitions/lqmtools.json"
    fileDestination     = ""
    incrementFile       = true

#Toolchain configuration
[[ToolChains]]
    active = true
    name   = "anl-flextext-test"
    chain  = [ "flextext-tool" ]
"""



import csv
import logging
import os

from lqmt.lqm.data import AlertAction
from lqmt.lqm.parsers.FlexTransform.parser import FlexTransformParser
from lqmt.lqm.tool import Tool


class ToFlexText(Tool):
    def __init__(self, config):
        """
        ToFlexText tool. Used to reformat CTI data in a user configured manner.

        :param config: configuration file
        """
        super().__init__(config, [AlertAction.get('All')])
        self._logger = logging.getLogger("LQMT.FlexText.{0}".format(self.getName()))

        # initialize parser with a dict created with variables from flextext configuration
        self._parser = FlexTransformParser({'CSV': self._config.flext_config})
        for file_type, source_config in self._config.source_configs.items():
            self._parser.add_parser(file_type, source_config)

        self._file_obj = None
        self._processed = []

    def initialize(self):
        super().initialize()

    def process(self, datafile, meta):
        """
        Process function. Handles the processing of data for the tool. Does so by calling the FlexText parser

        :param datafile: String that contains the path to the alert being processed.
        :param meta: meta data of the datafile. Used to assign correct parser
        """
        if self._file_obj is None and not self.openfile():
            self.disable()
        if self.isEnabled():
            if datafile not in self._processed:
                # Use flextext parser to parse datafile.
                self._parser.parseflextext(datafile, meta, self._file_obj, self._config.config_to_str())
                self._processed.append(datafile)

    def openfile(self):
        """
        Creates and opens the file specified in the user configuration, specifically the file_destination variable.
        The file object create here is then passed to the parser
        """
        try:
            file = self._config.file_destination
            file_dir = os.path.dirname(file)
            if not os.path.exists(file_dir):
                os.makedirs(file_dir, 0o755, True)
            self._file_obj = open(file, 'a', newline='')
            self.writeheader()
            return True
        except Exception as e:
            self._logger.error("Unable to open csv file: {0}".format(self._config.file_destination))
            self._logger.error(e)
            return False

    def writeheader(self):
        """
        When inserting the header using FlexT, the header gets repeated everytime a file is processed due to the way the
        config is currently sent over. This (hacky) solution writes the header to the file before any configuration data
        is sent over to FlexT. After the header is written the header_line value is set to False so that the header
        isn't rewritten when the configuration is passed to FlexT.
        """

        if self._config.header_line:
            quote_style = None
            if self._config.quote_style.lower() == 'none':
                quote_style = csv.QUOTE_NONE
            elif self._config.quote_style.lower() == 'nonnumeric':
                quote_style = csv.QUOTE_NONNUMERIC
            elif self._config.quote_style.lower() == 'all':
                quote_style = csv.QUOTE_ALL
            elif self._config.quote_style.lower() == 'minimal':
                quote_style = csv.QUOTE_MINIMAL

            csv.register_dialect('flext',
                                 delimiter=self._config.delimiter,
                                 quotechar=self._config.quote_char,
                                 escapechar=bytes(self._config.escape_char, "utf-8").decode("unicode_escape"),
                                 doublequote=self._config.double_quote,
                                 lineterminator='\r\n',
                                 quoting=quote_style
                                 )

            writer = csv.DictWriter(self._file_obj, self._config.fields, dialect='flext')

            writer.writeheader()
            self._config.header_line = False

    def commit(self):
        pass

    def cleanup(self):
        if self._file_obj is not None:
            self._file_obj.close()



import os
import time
import uuid
from unittest import TestCase, main
from lqmt.lqm.parsers.STIXParser.parser import STIXParser
from lqmt.lqm.systemconfig import SystemConfig
from lqmt.test.test_data.sample_inputs import STIXRULES, YARAEXAMPLE


class TestSTIXParser(TestCase):

    def setUp(self):
        sysconf = SystemConfig()
        self.sys_config = sysconf.getConfig()
        self.sys_config = self.sys_config['parsers']
        self.stixparser = STIXParser()
        self.time = str(time.time()).split('.')[0]
        self.directory = os.path.dirname(__file__)
        self.test_file = self.directory + "/" + STIXRULES

    def getMeta(self):
        """
        Method for generating metadata. Relatively static for now, but can be expanded to be randomized on each test run

        :return: Returns metadata in a dictionary
        """
        meta = {
            "PayloadFormat": 'STIX',
            "SendingSite": "Test Site",
            "PayloadType": "Alert",
            "UploadID": str(uuid.uuid4()).upper(),
            "FileName": "TestAlert",
            "SentTimestamp": self.time,
            "DownloadElementExtendedAttribute": [
                {
                    "Field": "Custodian",
                    "Value": "Test Custodian"
                },
                {
                    "Field": "Markings",
                    "Value": "TLP"
                },
                {
                    "Field": "Originator",
                    "Value": "Test Originator"
                },
                {
                    "Field": "Version",
                    "Value": "1.1.1"
                }
            ]
        }

        return meta

    def __default_config(self):
        # setup config context for test
        self.stixparser._sources = []
        self.stixparser._elements = ['indicators']
        self.stixparser._rules = ['snort', 'yara']

    def test_stix_no_sources(self):
        # Test runs with no source filtering to confirm parsed
        self.__default_config()

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertIsNotNone(alerts[0]._full_rules)
        self.assertIsNotNone(alerts[0]._rawfile)
        self.assertEquals(len(alerts[0]._stix_elements[0]), 2)

    def test_stix_source1(self):
        # Test - find Information_Source
        self.__default_config()
        self.stixparser._sources = ['Test Source']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertIsNotNone(alerts[0]._full_rules)
        self.assertIsNotNone(alerts[0]._rawfile)
        self.assertEquals(len(alerts[0]._stix_elements[0]), 2)

    def test_stix_source2(self):
        # Test - find Custodian
        self.__default_config()
        self.stixparser._sources = ['Test Custodian']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertIsNotNone(alerts[0]._full_rules)
        self.assertIsNotNone(alerts[0]._rawfile)
        self.assertEquals(len(alerts[0]._stix_elements[0]), 2)

    def test_stix_source3(self):
        # Test - find Originator
        self.__default_config()
        self.stixparser._sources = ['Test Originator']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertIsNotNone(alerts[0]._full_rules)
        self.assertIsNotNone(alerts[0]._rawfile)
        self.assertEquals(len(alerts[0]._stix_elements[0]), 2)

    def test_stix_source4(self):
        # Test - Source not found
        self.__default_config()
        self.stixparser._sources = ['Not Found']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser returned empty list
        self.assertEquals(alerts, [])

    def test_stix_source5(self):
        # Test - find multiple sources
        self.__default_config()
        self.stixparser._sources = ['Test Originator', 'Not Found']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertIsNotNone(alerts[0]._full_rules)
        self.assertIsNotNone(alerts[0]._rawfile)
        self.assertEquals(len(alerts[0]._stix_elements[0]), 2)

    def test_stix_elements1(self):
        # Test - find Originator
        self.__default_config()
        self.stixparser._elements = ['StIx_header', 'Indicators']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # confirm Parser filled all data variables, and found snort and yara rules
        self.assertIsNotNone(alerts)
        self.assertIsNotNone(alerts[0]._stix_elements)
        self.assertEquals(len(alerts[0]._stix_elements), 2)

    def test_empty_rules(self):
        self.__default_config()
        self.stixparser._elements = []
        self.stixparser._rules = []

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # ensure empty full context & empty rules dict
        self.assertEquals(alerts[0]._rules, {})
        self.assertEquals(alerts[0]._full_rules, [])

    def test_snort_rules(self):
        self.__default_config()
        self.stixparser._elements = []
        self.stixparser._rules = ['snort']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        self.assertEquals(len(alerts[0]._rules['snort']), 2)  # snort entry size 2
        self.assertEquals(alerts[0]._rules['snort'][0], 'alert tcp any any -> any any (msg:"FOX-SRT - Flowbit - TLS-SSL Client Hello"; flow:established; dsize:< 500; content:"|16 03|"; depth:2; byte_test:1, <=, 2, 3; byte_test:1, !=, 2, 1; content:"|01|"; offset:5; depth:1; content:"|03|"; offset:9; byte_test:1, <=, 3, 10; byte_test:1, !=, 2, 9; content:"|00 0f 00|"; flowbits:set,foxsslsession; flowbits:noalert; threshold:type limit, track by_src, count 1, seconds 60; reference:cve,2014-0160; classtype:bad-unknown; sid: 21001130; rev:9;)')
        self.assertEquals(alerts[0]._rules['snort'][1], 'alert tcp any any -> any any (msg:"FOX-SRT - Suspicious - TLS-SSL Large Heartbeat Response"; flow:established; flowbits:isset,foxsslsession; content:"|18 03|"; depth: 2; byte_test:1, <=, 3, 2; byte_test:1, !=, 2, 1; byte_test:2, >, 200, 3; threshold:type limit, track by_src, count 1, seconds 600; reference:cve,2014-0160; classtype:bad-unknown; sid: 21001131; rev:5;)')
        self.assertNotIn('yara', alerts[0]._rules)  # yara entry is empty

    def test_yara_rules(self):
        self.__default_config()
        self.stixparser._elements = []
        self.stixparser._rules = ['yara']

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        self.assertEquals(len(alerts[0]._rules['yara']), 1)  # snort entry size 2
        self.assertEquals(alerts[0]._rules['yara'][0], YARAEXAMPLE)
        self.assertNotIn('snort', alerts[0]._rules)  # yara entry is empty

    def test_full_context(self):
        self.__default_config()
        self.stixparser._elements = []

        alerts = self.stixparser.parse(self.test_file, self.getMeta())

        # if xsi:type is in the entry, it grabbed the full context
        self.assertEquals(len(alerts[0]._full_rules), 2)
        self.assertEquals(alerts[0]._full_rules[0]['xsi:type'], 'snortTM:SnortTestMechanismType')
        self.assertEquals(alerts[0]._full_rules[1]['xsi:type'], 'yaraTM:YaraTestMechanismType')


if __name__ == '__main__':
    main()



import ast
import logging
from lqmt.lqm.logging import LQMLogging
from .config import LQMToolConfig


# based on filename, place file either in the metafiles dict or the datafiles list
# NOTE: Assumes metafiles begin with . and datafiles do not.

class LQMToolController:
    def __init__(self, configfile):
        """
        Controller for LQMT.
        :param configfile: User configuration file
        """

        self._logger = logging.getLogger("LQMT.Controller")
        self._logger.info("Starting LQMTool")

        self._config = LQMToolConfig(configfile)
        self.toolChains = self._config.getToolChains()
        self.numAlerts = 0
        self.src = None

    def run(self):
        """
        Main function of the controller. Runs through the various methods used to gather alert files, parse them,
        and send the parsed alert data to the various tools.
        """
        self.pull()
        self.push()

    def push(self):
        """
        Function for initializing and running all the user defined "from" tools.
        """
        # TODO: if no push tools configured, do we still want to process?
        if not self.toolChains['push']:
            return

        alert_files = self._initialize()
        if alert_files:
            for data, unparsed_metadata in alert_files:
                metadata = self._parsemeta(unparsed_metadata)
                if metadata:
                    filters = self._config.getSourceFilters()
                    if filters:
                        if filters.checkAllFilters(metadata):
                            self._parse(data, metadata)
                    else:
                        self._parse(data, metadata)
        self._chainCleanup()

    def pull(self):
        # TODO: All tool functions contained to this function. Should give pull tools their own chain type and break
        # out the other functions out to the chain class. Similar to how it's done for push tools now.
        for chain in self.toolChains['pull']:
            chain.pull_process()

    def _parsemeta(self, metafile):
        """
        Parses metadata files
        :param metafile: Path to metadata file
        :return: Returns parsed metadata if the file path is valid. If not, then it returns None
        """
        try:
            f = open(metafile, 'r')
            meta = ast.literal_eval(f.read())
            f.close()
            return meta
        except Exception as inst:
            self._logger.error('An exception occurred while opening/parsing {0}:'.format(metafile))
            self._logger.error(str(inst))
            return None

    def _initialize(self):
        """
        Initializes LQMT. This includes initializing toolchains and their respective tools, and then getting all alert
        files specified from sources defined in the user configuration.
        :return: Returns filesToProcess object, which is defined in sourcedir.py. It's an object that has a custom
        __iter__ method that is used to traverse the top level alert directories provided by the user in user config
        """

        filesToProcess = None

        for chain in self.toolChains['push']:
            if chain._enabled:
                chain.initialize()
            else:
                self._logger.error("Toolchain '{0}' is disabled due to user configuration, or because no tools were "
                                   "correctly configured for the toolchain.".format(chain.getName()))
            chain.updateEnabled()

        if self._config.getSources():
            for src in self._config.getSources():
                self.src = src
                filesToProcess = src.getFilesToProcess()

        return filesToProcess

    def _parse(self, data, metadata):
        """
        Defines all parsers needed based on the metadata given. Once defined, the parsers are used to parse alert data
        and pass the data to tools.
        :param data: Alert data
        :param metadata: Alert metadata
        """

        parser = self._config.getParser(metadata["PayloadFormat"])

        try:
            if parser is not None:
                # tell each chain there is a new file
                alerts = parser.parse(data, metadata)
                if alerts:
                    for chain in self.toolChains['push']:
                        if chain.isEnabled():
                            chain.fileBegin()
                            self._process_alerts(alerts, chain, data, metadata)
                            self.src.processed(data)
                else:
                    self._logger.error("Processing error occurred. No processed alert data returned to LQMT.")
        except Exception as e:
            msg = "An exception occurred while processing file '{0}'".format(data)

            if not LQMLogging.isDebug():
                self._logger.error(msg)
            else:
                self._logger.exception(msg)
            self._logger.error(str(e))

    def _process_alerts(self, alerts, chain, datafile, metadata):
        """
        Takes alert data and passes it to active toolchains and their respective tools.
        :param alerts: Alert data
        :param chain: Toolchain
        """

        for alert in alerts:
            self.numAlerts += 1
            isWL = alert.isWhitelisted(self._config.getWhitelist())
            chain.process(alert, isWL, datafile, metadata)
            chain.fileDone()

    def _chainCleanup(self):
        """
        Cleans up tools that are done processing and logs statistics on amount of processed alerts.
        """

        for chain in self.toolChains['push']:
            if chain.isEnabled():
                chain.commit()
                chain.cleanup()

        for src in self._config.getSources():
            src.logStatistics(self.numAlerts)